# Data processing imports
from pyspark.sql import SparkSession
from pyspark import SparkContext, SparkConf
from datahandler import DataAggregator
from pyspark.ml import Pipeline
from pyspark.ml.feature import MinMaxScaler, VectorAssembler
from pyspark.ml.functions import vector_to_array
import pandas as pd

# Utils imports
from functools import reduce
from pathlib import Path
import os

# Imports for Google Colab Only
import findspark

findspark.init()



################################################################
################ LOAD and AGGREGATE the DATA ###################
#################################################################

def get_agg_data(target, local_csv_name='agg_df', data_path='data/input_data/'):
    # set Paths
    names_path = Path.cwd().joinpath(f"data/target_names.csv")
    agg_path = Path.cwd().joinpath(f"data/{local_csv_name}")

    if not agg_path.exists():

        # set up Spark Config
        conf = SparkConf().set('spark.ui.port', '4050')
        sc = SparkContext(conf=conf)
        spark = SparkSession \
            .builder \
            .appName("Spark_Aggregation") \
            .master("local[*]") \
            .getOrCreate()

        spark.conf.set('spark.sql.broadcastTimeout', 36000)
        spark.conf.set('spark.sql.debug.maxToStringFields', 2000)

        # Read data
        # data_path = '../../data/input_data/'
        file_names = [os.path.join(data_path, f) for f in os.listdir(data_path)]
        df = spark.read.options(header='True', inferSchema='True') \
            .option('timestampFormat', 'yyyy-MM-dd') \
            .csv(','.join(file_names))

        target_dtype = [dtype[1] for dtype in df.dtypes if dtype[0] == target][0]

        # Instantiate data aggregator
        aggregator = DataAggregator(df, target, target_dtype)

        # Set up the columns
        fix_cols = ['user_id', 'date', 'shop_name', 'shop_id', 'price']
        comp_cols = [col for col in aggregator.df.columns if 'company' in col]
        name_cols = [col for col in aggregator.df.columns if 'name' in col]
        cat_cols = [col for col in aggregator.df.columns if 'category' in col]

        from pyspark.sql.functions import col, countDistinct
        import pyspark.sql.functions as f

        df_count = aggregator.df.drop(*(fix_cols + comp_cols + cat_cols + name_cols))
        df_nunique = df_count.agg(*(countDistinct(col(c)).alias(c) for c in df_count.columns))
        continuous_cols = [col for col in df_nunique.columns if
                           df_nunique.select(f.collect_list(col)).first()[0][0] > 20]
        ratio_cols = [col for col in df_nunique.columns if col not in continuous_cols]

        # Aggregate data to make features
        features = []
        features.append(aggregator.make_price())
        features.append(aggregator.make_week_sales())
        features.append(aggregator.make_season_sales())

        if target == 'user_id':
            features.append(aggregator.make_count(col='shops', by='shop_id'))
        else:
            features.append(aggregator.make_count(col='clients', by='user_id'))
            features.append(aggregator.make_rating(lims=(1, 5)))
            if 'company_id' in aggregator.df.columns:
                features.append(aggregator.make_count(col='shops', by='shop_id'))

        for c in ratio_cols:
            features.append(aggregator.make_ratios(by='user_id', feat=c, prefix=c[:3].upper()))

        for c in cat_cols:
            features.append(aggregator.make_ratios(by='shop_id', feat=c, prefix=c[:3].upper()))

        for c in continuous_cols:
            features.append(aggregator.make_continuous(c))

        # Merge all features into a Dataset
        agg_df = reduce(lambda l, r: l.join(r, on=target, how='inner'), features)

        # Execute function as a Spark vectorized UDF
        feat_cols = [feat for feat in agg_df.schema.names if feat != target]
        assembler = VectorAssembler(inputCols=feat_cols, outputCol="features")
        scaler = MinMaxScaler(inputCol="features", outputCol="features_scaled")
        pipeline = Pipeline(stages=[assembler, scaler])
        scalerModel = pipeline.fit(agg_df)
        scaled_agg_df = scalerModel.transform(agg_df)
        scaled_agg_df = scaled_agg_df.withColumn("scale", vector_to_array("features_scaled")) \
            .select([target] + [col("scale")[i] for i in range(len(feat_cols))])

        for i in range(len(feat_cols)):
            scaled_agg_df = scaled_agg_df.withColumnRenamed(f"scale[{i}]", feat_cols[i])

        scaled_agg_df.write.option("header", True).csv(agg_path.as_posix())
        print('     Successfully Saved Aggregated Data!')
    else:
        print('     Aggregated Data Already Exists!')

    print('         Loading...')
    # Load SPARK processed data into a pandas Frame
    csv_files = [file for file in os.listdir(agg_path) if file.endswith("csv")]
    final_agg_df = []
    for csv_file in csv_files:
        final_agg_df.append(pd.read_csv(agg_path / csv_file))
    final_agg_df = pd.concat(final_agg_df)

    return df, final_agg_df
