import pandas as pd #
import numpy as np 
from functools import reduce
from sklearn.preprocessing import MinMaxScaler as mms #
pd.options.mode.chained_assignment = None


class DataAggregator():

  def __init__(self, df, target, target_dtype):
    self.target = target
    self.df = df
    self.target_ids = [row[0] for row in self.df.select(self.target).distinct().collect()]
    self.target_dtype = target_dtype

  @staticmethod
  def quantile_denoise(df, col, z=1.5):
    q1=df[col].quantile(.25); q3=df[col].quantile(.75); iqr=q3-q1
    #mask = df[col].between(q1-z*iqr, q3+z*iqr, inclusive=True)
    mask = df[col].between(q1-z*iqr, q3+z*iqr, inclusive="both")
    return df.loc[mask,:]

  def make_rating(self, lims=(1,5)):
      print('Making Rating feature...')
      target = self.target
      target_dtype = self.target_dtype 
      quantile_denoise = self.quantile_denoise
      schema = f"{target} {target_dtype}, "+ \
               f"rating double"

      def agg_func(df:pd.DataFrame) -> pd.DataFrame:
        # Make ratings column from price
        c_dfs = []
        c_dfs_append = c_dfs.append

        # Using proper Columns
        cols = [target,'price']
        if target!='card_id': cols.append('card_id')
    
        # Get cancelation ratings 
        if (df['price'] < 0).any():
            neg_df = df.loc[df.price<0,cols]
            neg_df['rating'] = 0
            c_dfs_append(neg_df)

        # Get Purchase ratings
        c_df = df.loc[df.price>0,cols]
        shape = c_df.shape[0]
        if shape == 0: #0 priced items
            c_df = df.copy() 
            c_df['rating'] = np.nan
        elif shape == 1: 
            c_df['rating'] = (lims[0] + lims[1])/2  
        else:
            c_df = quantile_denoise(c_df, col='price')
            scaler = mms(feature_range=lims)
            c_df['rating'] = scaler.fit_transform(X=c_df['price'].values.reshape(-1,1)) 
        c_dfs_append(c_df)

        std_pr_df = pd.concat(c_dfs)
        ratings = std_pr_df.groupby([target]).mean()[['rating']]
        ratings = ratings.reset_index().rename(columns={'index':target})
        return ratings


      ratings = self.df.groupby(target) \
                   .applyInPandas(agg_func, schema=schema)
      ratings = ratings.filter(ratings.rating.isNotNull())
      print('         Succesfully Made!')
      return ratings


  def make_price(self):
      print('Making Price feature...')
      #from pyspark.sql.functions import pandas_udf, PandasUDFType
      #from pyspark.sql.types import StructType, StructField, IntegerType, DoubleType

      target = self.target
      target_dtype = self.target_dtype 
      quantile_denoise = self.quantile_denoise
      pos_den_df = self.df.filter(self.df.price > 0) 
      neg_den_df = self.df.filter(self.df.price < 0) 
      pos_col, neg_col = 'purchase', 'cancel'

      def get_schema(col, target=target):
        # schema = StructType([StructField(target, IntegerType(), True),
        #           StructField(f"{col}_mean", DoubleType(), True),
        #           StructField(f"{col}_ptp", IntegerType(), True),
        #           StructField(f"{col}_size", IntegerType(), True)])
        schema = f"{target} {target_dtype}, "+ \
                f"{col}_mean double, "+ \
                f"{col}_ptp integer, "+ \
                f"{col}_size integer"
        return schema

      def function_generator(col):
        #schema = get_schema(col)
        #@pandas_udf(schema, PandasUDFType.GROUPED_MAP)
        def agg_func(df: pd.DataFrame) -> pd.DataFrame:

          #Denoise the dataset
          if target == 'card_id': pass
          else:
            shape = df.shape[0]
            if shape < 4: pass
            else: df = quantile_denoise(df, col='price')

          #Define dictionary to be turned into pd.DataFrame
          row = {
              target: df[target].unique()[0],
              f"{col}_mean": [df['price'].mean()],
              f"{col}_ptp": [df['price'].max() - df['price'].min()],
              f"{col}_size": [df.shape[0]]
          }
          agg_df = pd.DataFrame(row)
          return agg_df  
        return agg_func      

      pos_agg = function_generator(pos_col)
      pur_df = pos_den_df.select('price',target) \
                       .groupby(target) \
                       .applyInPandas(pos_agg, schema=get_schema(pos_col))
                       #.apply(pos_agg)
      
      if neg_den_df.count() == 0:
        print('         Purchase Succesfully Made!')
        return pur_df
      else:
        neg_agg = function_generator(neg_col)
        ccl_df = neg_den_df.select('price',target) \
                       .groupby(target) \
                       .applyInPandas(neg_agg, schema=get_schema(neg_col))
                       #.apply(neg_agg)
        pr_df = pur_df.join(ccl_df, on=target, how='outer')
        pr_df = pr_df.na.fill(value=0) #in case there are no cancel in some targets
        print('         Purchase & Cancelation Succesfully Made!')
        return pr_df
        

  def make_continuous(self, feat):
      print(f'Making {feat} feature...')
      target = self.target
      target_dtype = self.target_dtype 
      quantile_denoise = self.quantile_denoise

      schema = f"{target} {target_dtype}, "+ \
                f"{feat}_mean double, "+ \
                f"{feat}_ptp integer"

      def agg_func(df: pd.DataFrame) -> pd.DataFrame:
        #feat_den_df = quantile_denoise(df, col=feat)
        if df['card_id'].nunique() == 1: 
          feat_df = df[[target,feat]].groupby(target) \
                        .aggregate([np.mean, np.ptp])
        else:
          feat_df = df[[feat, target, 'card_id']] \
                          .groupby([target, 'card_id']) \
                          .mean()[[feat]]
          feat_df = feat_df.reset_index() \
                          .drop(['card_id'], axis=1) \
                          .groupby(target) \
                          .aggregate([np.mean, np.ptp])
        feat_df = feat_df.reset_index() \
                          .rename(columns={f"'('{feat}', 'mean')'": f"{feat}_mean",
                                           f"'('{feat}', 'ptp')'": f"{feat}_ptp"})     
        return feat_df

      feat_df = self.df.groupby(target) \
                    .applyInPandas(agg_func, schema=schema)
      print('         Succesfully Made!')
      return feat_df

  def make_week_sales(self):
    print('Making Week Sales feature...')
    target = self.target
    target_dtype = self.target_dtype 
    date_df = self.df.select(target,'date','price')
    schema = f"{target} {target_dtype}, "+ \
              "weekday_ratio double, "+ \
              "weekend_ratio double"

    def agg_func(df: pd.DataFrame) -> pd.DataFrame:
      df['date'] = pd.to_datetime(df['date'])
      w_dfs = []
      w_dfs_append = w_dfs.append
      w_dict = {'weekday':[0,1,2,3,4], 'weekend':[5,6]}
      for we in w_dict.keys():
        indexer = (df.date.dt.weekday.isin(w_dict[we]))&(df.price>0)
        w_df = df[indexer][target].value_counts().to_frame()
        w_df /= len(w_dict[we])
        w_df.rename(columns={target:f'{we}_ratio'}, inplace=True)
        w_dfs_append(w_df)
      we_df = reduce(lambda left,right: pd.merge(left, right, how='outer', 
                                  right_index=True, left_index=True), w_dfs)
      we_df['sum'] = we_df.sum(axis=1)
      we_df = we_df.iloc[:,:-1].div(we_df['sum'], axis=0)
      we_df.fillna(0, inplace=True)
      we_df = we_df.reset_index().rename(columns={'index':target})
      return we_df

    we_df = date_df.groupby(target) \
                   .applyInPandas(agg_func, schema=schema)
    print('         Succesfully Made!')
    return we_df

  def make_season_sales(self):
    print('Making Season Sales feature...')
    target = self.target
    target_dtype = self.target_dtype 
    date_df = self.df.select(target,'date','price')
    schema = f"{target} {target_dtype}, "+ \
              "Season1_ratio double, "+ \
              "Season2_ratio double, "+ \
              "Season3_ratio double, "+ \
              "Season4_ratio double"

    def agg_func(df: pd.DataFrame) -> pd.DataFrame:
      s_dfs = []
      s_dfs_append = s_dfs.append
      df['date'] = pd.to_datetime(df['date'])
      for i, s in enumerate(list(range(1,13))[::3], 1): #first month of each season
        indexer = (df.date.dt.month.isin(range(s,s+3)))&(df.price>0) 
        s_dfs_append(df[indexer][target].value_counts().to_frame() \
                                .rename(columns={target:f'Season{i}_ratio'}))
      sa_df = reduce(lambda left,right: pd.merge(left, right, how='outer', 
                                right_index=True, left_index=True), s_dfs)
      sa_df['sum'] = sa_df.sum(axis=1)
      sa_df = sa_df.iloc[:,:-1].div(sa_df['sum'], axis=0)
      sa_df.fillna(0, inplace=True) #0 means no sales occured during that period
      sa_df = sa_df.reset_index().rename(columns={'index':target})
      return sa_df

    sa_df = date_df.groupby(target) \
                   .applyInPandas(agg_func, schema=schema)
    print('         Succesfully Made!')
    return sa_df

  def make_count(self, col, by): 
    print(f'Making {col} Count feature...')
    #col = {'shops', 'clients'}
    #by = {'item_id', 'card_id'}
    target = self.target
    target_dtype = self.target_dtype 
    schema = f"{target} {target_dtype}, "+ \
             f"{col} integer"

    def agg_func(df: pd.DataFrame) -> pd.DataFrame:
      if (target=='item_id')&(by=='item_id'):
        item_comp = df[['company_id', by]].drop_duplicates().set_index(by)
        count_df = item_comp['company_id'].value_counts().to_frame()
        item = item_comp.index[0]
        index = item_comp.loc[item,'company_id']
        item_comp.loc[item,col] = count_df.loc[index,'company_id']
        item_comp = item_comp.reset_index().rename(columns={'index':target})
        item_comp = item_comp.drop(['company_id'], axis=1)
        return item_comp
      else:
        count_df = df[[target, by]].drop_duplicates().groupby(target).count()
        count_df.columns = [col]
        count_df = count_df.reset_index().rename(columns={'index':target})
        return count_df

    cnt_df = self.df.groupby(target) \
                   .applyInPandas(agg_func, schema=schema)
    print('         Succesfully Made!')
    return cnt_df
  
  def make_ratios(self, by, feat, prefix):
    import pyspark.sql.functions as f
    import json

    def isEnglish(s):
      try:
          s.encode('ascii')
      except UnicodeEncodeError:
          return False
      else:
          return True

    print(f'Making {feat} feature...')
    target = self.target
    target_dtype = self.target_dtype 
    unique_feat_df = self.df.select(feat).distinct()
    unique_feat = unique_feat_df.select(f.collect_list(feat)).first()[0]
    for col in unique_feat:
      if isEnglish(str(col)):
        cols_dict = None
      else:
        cols_dict = {col:i for i,col in zip(range(len(unique_feat)), unique_feat)}
        break
    if cols_dict is not None:
      pivot_cols = [prefix+'_'+str(cols_dict[col]) for col in unique_feat]
      json = json.dumps(cols_dict)
      f = open(f"data/{feat}_dict.json","w")
      f.write(json)
      f.close()
    else:
      pivot_cols = [prefix+'_'+str(col) for col in unique_feat]
    schema_cols = [f"{target} {target_dtype}"]
    schema_cols.extend([col+' double' for col in pivot_cols])
    schema = ', '.join(schema_cols)
    
    cols = list(set([by, target, feat]))

    def agg_func(df: pd.DataFrame) -> pd.DataFrame:
      df_sp = df[cols].drop_duplicates()#.set_index(self.target)
      cnt_df = df_sp[feat].value_counts(normalize=True)
      agg_df = pd.DataFrame([[0]*len(unique_feat)], columns=unique_feat)
      for col in cnt_df.index:
        agg_df[col] = cnt_df[col]
      agg_df.rename(columns=dict(zip(agg_df.columns,pivot_cols)), inplace=True)
      agg_df[target] = df[target].unique()[0]
      return agg_df

    df_sp = self.df.groupby(target) \
                   .applyInPandas(agg_func, schema=schema)
    print('         Succesfully Made!')
    return df_sp

    
