import torch
import torch.nn as nn
from torch.nn import Parameter
import torch.optim as optim
from torch.autograd import Variable

import numpy as np
import os
import math
import collections
import json

from sklearn.metrics.cluster import normalized_mutual_info_score
from sklearn.cluster import KMeans
import pandas as pd
from pathlib import Path

def buildNetwork(layers, activation="relu", dropout=0):
    net = []
    for i in range(1, len(layers)):
        net.append(nn.Linear(layers[i-1], layers[i]))
        if activation=="relu":
            net.append(nn.ReLU())
        elif activation=="sigmoid":
            net.append(nn.Sigmoid())
        if dropout > 0:
            net.append(nn.Dropout(dropout))
    return nn.Sequential(*net)

def acc(y_true, y_pred):
    # This require sklearn installed
    y_true = y_true.astype(np.int64)
    assert y_pred.size == y_true.size
    D = max(y_pred.max(), y_true.max()) + 1
    w = np.zeros((D, D), dtype=np.int64)
    for i in range(y_pred.size):
        w[y_pred[i], y_true[i]] += 1
    # from sklearn.utils.linear_assignment_ import linear_assignment
    from scipy.optimize import linear_sum_assignment as linear_assignment
    ind = linear_assignment(w.max() - w)
    return sum([w[i, j] for i, j in ind]) * 1.0 / y_pred.size


class IDEC(nn.Module):
    def __init__(self, input_dim=784, z_dim=10, n_clusters=10, mu_param=True,
        encodeLayer=[400], decodeLayer=[400], activation="relu", dropout=0, alpha=1., gamma=0.1):
        super(self.__class__, self).__init__()
        self.z_dim = z_dim
        self.layers = [input_dim] + encodeLayer + [z_dim]
        self.activation = activation
        self.dropout = dropout
        self.encoder = buildNetwork([input_dim] + encodeLayer, activation=activation, dropout=dropout)
        self.decoder = buildNetwork([z_dim] + decodeLayer, activation=activation, dropout=dropout)
        self._enc_mu = nn.Linear(encodeLayer[-1], z_dim)
        self._dec = nn.Linear(decodeLayer[-1], input_dim)

        self.n_clusters = n_clusters
        self.alpha = alpha
        self.gamma = gamma
        self.is_available_to_cuda = 'cuda' if torch.cuda.is_available() else 'cpu'
        if mu_param:
          self.mu = Parameter(torch.Tensor(n_clusters, z_dim))
        else:
          self.mu = torch.Tensor(n_clusters, z_dim).to(self.is_available_to_cuda)

    def save_model(self, path):
        torch.save(self.state_dict(), path)

    def load_model(self, path):
        pretrained_dict = torch.load(path, map_location=lambda storage, loc: storage)
        model_dict = self.state_dict()
        pretrained_dict = {k: v for k, v in pretrained_dict.items() if k in model_dict}
        model_dict.update(pretrained_dict) 
        self.load_state_dict(model_dict)

    def forward(self, x):
        h = self.encoder(x)
        z = self._enc_mu(h)
        h = self.decoder(z)
        xrecon = self._dec(h)
        # compute q -> NxK
        q = self.soft_assign(z)
        return z, q, xrecon

    def soft_assign(self, z):
        q = 1.0 / (1.0 + torch.sum((z.unsqueeze(1) - self.mu)**2, dim=2) / self.alpha)
        q = q**(self.alpha+1.0)/2.0
        q = q / torch.sum(q, dim=1, keepdim=True)
        return q

    def encodeBatch(self, X, batch_size=8):
        use_cuda = torch.cuda.is_available()
        if use_cuda:
            self.cuda()
        
        encoded = []
        self.eval()
        num = X.shape[0]
        num_batch = int(math.ceil(1.0*X.shape[0]/batch_size))
        for batch_idx in range(num_batch):
            xbatch = X[batch_idx*batch_size : min((batch_idx+1)*batch_size, num)]
            inputs = Variable(xbatch)
            z,_, _ = self.forward(inputs)
            encoded.append(z.data)

        encoded = torch.cat(encoded, dim=0)
        return encoded

    def cluster_loss(self, p, q):
        def kld(target, pred):
            return torch.mean(torch.sum(target*torch.log(target/(pred+1e-6)), dim=1))
        kldloss = kld(p, q)
        return self.gamma*kldloss

    def recon_loss(self, x, xrecon):
        recon_loss = torch.mean((xrecon-x)**2)
        return recon_loss

    # def global_size_loss(self, p, cons_detail):
    #     m_p = torch.mean(p, dim=0)
    #     m_p = m_p / torch.sum(m_p)
    #     return torch.sum((m_p-cons_detail)*(m_p-cons_detail))

    # def assignment_ct_loss(self, q, ct_mask):
    #     q_core = q*ct_mask
    #     q_tail = q*(1 - ct_mask)

    #     core_cnt = torch.mean(q_core, dim=0) / 0.4
    #     tail_cnt = torch.mean(q_tail, dim=0) / 0.6

    #     return torch.sum((core_cnt - tail_cnt)**2)

    def assignment_ct_loss(self, q, ct_mask, assign_size, data_size):
        q_core = q * ct_mask
        min_, max_ = assign_size[:2]  # 1, 2 #4500, 7500
        min_thresh, max_thresh = min_ / data_size, max_ / data_size  # min_/110629, max_/110629
        n_cls = q.shape[1]

        min_loss = (torch.mean(q_core, dim=0) - torch.Tensor([min_thresh] * n_cls).to(self.is_available_to_cuda))
        min_loss = torch.where(min_loss < 0, min_loss, torch.zeros(n_cls).to(self.is_available_to_cuda))

        max_loss = (torch.mean(q_core, dim=0) - torch.Tensor([max_thresh] * n_cls).to(self.is_available_to_cuda))
        max_loss = torch.where(max_loss < 0, torch.zeros(n_cls).to(self.is_available_to_cuda), max_loss)

        core_loss = torch.sum(min_loss ** 2 + max_loss ** 2)

        q_tail = q * (1 - ct_mask)
        min_, max_ = assign_size[2:]  # 3000, 8000 #4500, 7500
        min_thresh, max_thresh = min_ / data_size, max_ / data_size  # min_/110629, max_/110629
        n_cls = q.shape[1]

        min_loss = (torch.mean(q_tail, dim=0) - torch.Tensor([min_thresh] * n_cls).to(self.is_available_to_cuda))
        min_loss = torch.where(min_loss < 0, min_loss, torch.zeros(n_cls).to(self.is_available_to_cuda))

        max_loss = (torch.mean(q_tail, dim=0) - torch.Tensor([max_thresh] * n_cls).to(self.is_available_to_cuda))
        max_loss = torch.where(max_loss < 0, torch.zeros(n_cls).to(self.is_available_to_cuda), max_loss)

        tail_loss = torch.sum(min_loss ** 2 + max_loss ** 2)

        return core_loss + tail_loss

    def global_size_loss(self, p, size_tuple, data_size):
        min_, max_ = size_tuple  # 4, 8 #4500, 7500
        min_thresh, max_thresh = min_ / data_size, max_ / data_size  # min_/63, max_/63
        n_cls = p.shape[1]

        min_loss = (torch.mean(p, dim=0) - torch.Tensor([min_thresh] * n_cls).to(self.is_available_to_cuda))
        min_loss = torch.where(min_loss < 0, min_loss, torch.zeros(n_cls).to(self.is_available_to_cuda))

        max_loss = (torch.mean(p, dim=0) - torch.Tensor([max_thresh] * n_cls).to(self.is_available_to_cuda))
        max_loss = torch.where(max_loss < 0, torch.zeros(n_cls).to(self.is_available_to_cuda), max_loss)

        return torch.sum(min_loss ** 2 + max_loss ** 2)

    def target_distribution(self, q):
        p = q**2 / torch.sum(q, dim=0)
        p = p / torch.sum(p, dim=1, keepdim=True)
        return p
    
    def save_progress(self, save_path, epoch, latent, y_pred, y, intermediate_results):
        self.save_model(os.path.join(save_path,'model','save_model_%d.pt'%(epoch)))
        df = pd.DataFrame(latent.cpu().numpy())
        df["y"] = y
        df["y_pred"] = y_pred
        df.to_pickle(os.path.join(save_path,"enc_pred_%d.pkl"%(epoch)))

        cls_centers = pd.DataFrame(self.mu.data.cpu().numpy())
        cls_centers.to_pickle(os.path.join(save_path,"cluster_centers_%d.pkl"%(epoch)))

        if y is not None:
          intermediate_results["acc"][str(epoch)] = acc(y, y_pred)
          intermediate_results["nmi"][str(epoch)] = normalized_mutual_info_score(y, y_pred)
          with open(os.path.join(save_path,"intermediate_results.json"), "w") as fp:
              json.dump(intermediate_results, fp)

    def predict(self, X, y, save_path):
        use_cuda = torch.cuda.is_available()
        if use_cuda:
            self.cuda()
        latent = self.encodeBatch(X)
        q = self.soft_assign(latent)

        # evalute the clustering performance
        y_pred = torch.argmax(q, dim=1).data.cpu().numpy()
        y = y.data.cpu().numpy()
        df = pd.DataFrame(latent.cpu().numpy())
        df["y_pred"] = y_pred
        df['y'] = y
        df.to_pickle(os.path.join(save_path,"pred.pkl"))

        y = y.data.cpu().numpy()
        if y is not None:
            print("acc: %.5f, nmi: %.5f" % (acc(y, y_pred), normalized_mutual_info_score(y, y_pred)))
            final_acc = acc(y, y_pred)
            final_nmi = normalized_mutual_info_score(y, y_pred)
        return final_acc, final_nmi

    def fit(self, X, y, ct_mask, size_tuple, assign_tuple, target_type,
              lr, batch_size, num_epochs, update_interval, tol,
              centers_init, clustering_loss_weight=1, save_path=""):  

        # save intermediate results for plotting
        intermediate_results = collections.defaultdict(lambda:{})
        
        '''X: tensor data'''
        use_cuda = torch.cuda.is_available()
        if use_cuda:
            self.cuda()
        print("=====Training IDEC=======")
        optimizer = optim.Adam(filter(lambda p: p.requires_grad, self.parameters()), lr=lr)

        #if use_kmeans:
        if centers_init == "kmeans":
            print("Initializing cluster centers with kmeans.")
            kmeans = KMeans(self.n_clusters, n_init=20)
            data = self.encodeBatch(X)
            y_pred = kmeans.fit_predict(data.data.cpu().numpy())
            y_pred_last = y_pred
            self.mu.data.copy_(torch.Tensor(kmeans.cluster_centers_))
        elif centers_init == "random":
            # use kmeans to randomly initialize cluster ceters
            print("Randomly initializing cluster centers.")
            kmeans = KMeans(self.n_clusters, n_init=1, max_iter=1)
            data = self.encodeBatch(X)
            y_pred = kmeans.fit_predict(data.data.cpu().numpy())
            y_pred_last = y_pred
            self.mu.data.copy_(torch.Tensor(kmeans.cluster_centers_))
        elif centers_init == "custom":
            # use fixed user defined centers to guide the clustering
            print("Initializing cluster centers with custom user defined centers")
            cores_path = Path("./output/cores_only")
            cls_centr_files = [file for file in sorted(os.listdir(cores_path)) if "centers" in file]
            centers = pd.read_pickle(cores_path/cls_centr_files[-1]).values
            self.mu.data.copy_(torch.Tensor(centers))

        if y is not None:
            y = y.cpu().numpy()
            print("Kmeans acc: %.5f, nmi: %.5f" % (acc(y, y_pred), normalized_mutual_info_score(y, y_pred)))

        self.train()
        num = X.shape[0]
        num_batch = int(math.ceil(1.0*num/batch_size))

        final_acc, final_nmi, final_epoch = 0, 0, 0
        for epoch in range(num_epochs):
            if epoch%update_interval == 0:
                # update the targe distribution p
                latent = self.encodeBatch(X)
                q = self.soft_assign(latent)
                p = self.target_distribution(q).data

                # evalute the clustering performance
                y_pred = torch.argmax(q, dim=1).data.cpu().numpy()
                if size_tuple:
                    y_dict = collections.defaultdict(list)
                    ind1, ind2 = [], []
                    for i in range(y_pred.shape[0]):
                        y_dict[y_pred[i]].append(i)
                    for key in y_dict.keys():
                        if y is not None:
                            print("predicted class: ", key, " total: ", len(y_dict[key]))
                            #, " mapped index(ground truth): ", np.bincount(y[y_dict[key]]).argmax())

                if y is not None:
                    print("acc: %.5f, nmi: %.5f" % (acc(y, y_pred), normalized_mutual_info_score(y, y_pred)))
                    final_acc = acc(y, y_pred)
                    final_nmi = normalized_mutual_info_score(y, y_pred)
                    final_epoch = epoch

                # save model for plotting
                if save_path and (epoch%50 == 0 or epoch == num_epochs-1):
                    self.save_progress(save_path, epoch, latent, y_pred, y, intermediate_results)

                # check stop criterion
                stop_training = False
                try:
                    delta_label = np.sum(y_pred != y_pred_last).astype(np.float32) / num
                    y_pred_last = y_pred
                    if epoch>0 and delta_label < tol:
                        stop_training = True
                        print('delta_label ', delta_label, '< tol ', tol)
                        print("Reach tolerance threshold. Stopping training.")

                        # save model for plotting
                        if save_path:
                            self.save_progress(save_path, epoch, latent, 
                                               y_pred, y, intermediate_results)
                        break
                except:
                    pass

            if stop_training: break

            # train 1 epoch for clustering loss
            train_loss = 0.0
            recon_loss_val = 0.0
            cluster_loss_val = 0.0
            constraint_loss_val = 0.0
            for batch_idx in range(num_batch):
                xbatch = X[batch_idx*batch_size : min((batch_idx+1)*batch_size, num)]
                pbatch = p[batch_idx*batch_size : min((batch_idx+1)*batch_size, num)]
                ct_mask_batch = ct_mask[batch_idx*batch_size : min((batch_idx+1)*batch_size, num)] if assign_tuple else 0
                optimizer.zero_grad()
                inputs = Variable(xbatch)
                target = Variable(pbatch)
                z, qbatch, xrecon = self.forward(inputs)

                cluster_loss = self.cluster_loss(target, qbatch)
                recon_loss = self.recon_loss(inputs, xrecon)
                size_loss = self.global_size_loss(qbatch, size_tuple, num) if size_tuple else 0
                assign_loss = self.assignment_ct_loss(qbatch, ct_mask_batch, assign_tuple, num) if assign_tuple else 0
                constraint_loss = size_loss + assign_loss
                loss = cluster_loss + recon_loss + constraint_loss
                loss.backward()
                optimizer.step()
                cluster_loss_val += cluster_loss.data * len(inputs)
                recon_loss_val += recon_loss.data * len(inputs)
                constraint_loss_val += constraint_loss.data * len(inputs) 
                train_loss = clustering_loss_weight*cluster_loss_val + recon_loss_val + constraint_loss_val

            if constraint_loss_val != 0.0:
                print("#Epoch %3d: Total: %.4f Clustering Loss: %.4f Reconstruction Loss: %.4f Constraint Loss: %.4f"% (
                    epoch + 1, train_loss / num + constraint_loss_val/num_batch, cluster_loss_val / num, recon_loss_val / num, constraint_loss_val / num_batch))
            else:
                print("#Epoch %3d: Total: %.4f Clustering Loss: %.4f Reconstruction Loss: %.4f" % (
                    epoch + 1, train_loss / num, cluster_loss_val / num, recon_loss_val / num))

        return final_acc, final_nmi, final_epoch
