import torch.utils.data
import argparse
from dcaas.model import IDEC
from dcaas.agg_data import get_agg_data

import os
from pathlib import Path


def get_batch_size(size):
    sizes = [1e3, 1e4, 1e5, 1e6]
    batches = [16, 64, 128, 256]
    for s, b in zip(sizes, batches):
        if size < s:
            return b
    return 512


def get_min_max(size, n_clusters):
    pow = min([3, len(str(size // 10))])
    mod = 10 ** (pow - 1)

    init_min = int(size / n_clusters / 1.5)
    min_mod = init_min % mod
    min_s = init_min - min_mod if min_mod < mod / 2 else init_min + (mod - min_mod)

    init_max = int(size / n_clusters * 1.5)
    max_mod = init_max % mod
    max_s = init_max - max_mod if max_mod < mod / 2 else init_max + (mod - max_mod)
    return min_s, max_s


def train(agg_df, target_type, n_clusters, size_tuple, assign_tuple, batch_size, target, use_pretrain, lr, epochs, update_interval):
    # Model Parameters
    centers_init = "kmeans"
    mu_param = True
    ct_mask = None

    # Input data per use case
    if len(target_type) <= 2:
        if target_type == 'C':
            df = agg_df[agg_df.COR_T == 1]
            save_path = Path("./output/cores_only")
            (save_path / 'model').mkdir(parents=True, exist_ok=True)
        elif target_type == "T1":
            df = agg_df[agg_df.COR_T == 0]
            save_path = Path("./output/tails_only")
            (save_path / 'model').mkdir(parents=True, exist_ok=True)
        elif target_type == "T2":
            # This requires to run 'args.target_type=C' first
            df = agg_df[agg_df.COR_T == 0]
            centers_init = "custom"
            mu_param = False
            save_path = Path("./output/tails_to_cores")
            (save_path / 'model').mkdir(parents=True, exist_ok=True)
    else:
        df = agg_df
        ct = agg_df["COR_T"].values
        ct_mask = [[c] * n_clusters for c in ct]
        DEVICE_TYPE = 'cuda' if torch.cuda.is_available() else 'cpu'
        ct_mask = torch.Tensor(ct_mask).to(DEVICE_TYPE)
        save_path = Path("./output/tails_and_cores")
        (save_path / 'model').mkdir(parents=True, exist_ok=True)

    # Cluster size Constraints
    if size_tuple[0] < 0:
        size_tuple = None
    elif size_tuple[0] == 0:
        size = df.shape[0]
        size_tuple = get_min_max(size, n_clusters)
        print("Automatic Size Constraints are:")
        print(f"Cluster min size = {size_tuple[0]}")
        print(f"Cluster max size = {size_tuple[1]}")
    else:
        assert len(size_tuple) == 2, \
            "PLEASE ENTER 2 VALUES FOR [min] and [max] CLUSTER SIZE CONSTRAINTS"
        size_tuple = tuple(size_tuple)

    # Assignment size Constraints
    if assign_tuple[0] < 0:
        assign_tuple = None
    elif assign_tuple[0] == 0:
        print("Automatic Core and Tails Size Constraints are:")
        size = df["COR_T"].value_counts()
        c_size, t_size = size[1], size[0]
        c_assign_tuple = get_min_max(c_size, n_clusters)
        print(f"Core min size = {c_assign_tuple[0]}")
        print(f"Core max size = {c_assign_tuple[1]}")
        t_assign_tuple = get_min_max(t_size, n_clusters)
        print(f"Tail min size = {t_assign_tuple[0]}")
        print(f"Tail max size = {t_assign_tuple[1]}")
        assign_tuple = c_assign_tuple + t_assign_tuple
    else:
        assert len(assign_tuple) == 4, \
            """PLEASE ENTER 4 VALUES FOR [core_min], [core_max]
        and [tail_min], [tail_max] ASSIGNMENT SIZES CONSTRAINTS"""
        assign_tuple = tuple(assign_tuple)

    # Get model input Tensors and remaining Params
    df = df.drop(["COR_T", "COR_F"], axis=1)
    feat_cols = [feat for feat in df.columns if feat != target]
    X = df[feat_cols].values
    X = torch.Tensor(X).to(DEVICE_TYPE)
    y = None
    input_dim = X.shape[1]
    if batch_size == 0:
        batch_size = get_batch_size(df.shape[0])
    else:
        batch_size = batch_size
    print('batch_size = ', batch_size)
    zdim = n_clusters
    print("latent space dimension = ", zdim)

    # Initialize the model
    idec = IDEC(input_dim=input_dim, z_dim=zdim, n_clusters=n_clusters, mu_param=mu_param,
                encodeLayer=[500, 500, 2000], decodeLayer=[2000, 500, 500], activation="relu", dropout=0)

    # Print Network Structure
    print(idec)

    # Clustering data
    if use_pretrain:
        # Load pretrained model
        model_files = [file for file in sorted(os.listdir(save_path / 'model'))]
        pretrain_path = save_path / model_files[-1]
        idec.load_model(pretrain_path)

        # Predict Clusters
        acc, nmi = idec.predict(X, y, save_path=save_path)
        if y is not None:
            print("Accuracy:", acc)
            print("NMI;", nmi)
    else:
        # Train the network
        train_acc, train_nmi, epo = idec.fit(X, y, ct_mask, size_tuple, assign_tuple, target_type=target_type,
                                             lr=lr, batch_size=batch_size, num_epochs=epochs,
                                             save_path=save_path,
                                             update_interval=update_interval, tol=1 * 1e-3,
                                             centers_init=centers_init)
        # Report results
        print("Training Accuracy:", train_acc)
        print("Training NMI;", train_nmi)
        print("Training Epochs:", epo)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Constrained IDEC Parameters')
    parser.add_argument('--data_name', '-dn', type=str, default='agg_df',
                        help='Will be used as the folder name to store the SPARK aggregated data')
    parser.add_argument('--lr', type=float, default=0.001, metavar='N',
                        help='learning rate for training (default: 0.001)')
    parser.add_argument('--batch_size', '-bs', type=int, default=0, metavar='N',
                        help='input batch size for training (default: should be 256, but set to 0 for now)')
    parser.add_argument('--update_interval', type=int, default=1, metavar='N',
                        help='number of epochs to train (default: 1)')
    parser.add_argument('--epochs', type=int, default=200, metavar='N',
                        help='number of epochs to train (default: 200)')
    parser.add_argument('--use_pretrain', type=bool, default=False)
    parser.add_argument('--n_clusters', '-ncls', type=int, default=10)
    parser.add_argument('--target', '-t', type=str, default='user_id',
                        help='Target Instances to be Clustered (companies or shops or users)')
    parser.add_argument('--target_type', '-tt', type=str, default='ALL',
                        help="""C    : Cores only
                                T1   : Tails only
                                ALL  : Cores and Tails (simultaneously)
                                T2   : Tails (append Tails on Cores Centers)
                                       This require to run 'C' first """)
    parser.add_argument('--size_tuple', '-st', type=int, default=[-1], nargs="+",
                        help="""The min and max cluster size constraints:
                        -1    : Train w/o cluster size loss,
                        0     : Train with automatic cluster size constraints
                        (min, max) : Train with user defined cluster size constraints""")
    parser.add_argument('--assign_tuple', '-at', type=int, default=[-1], nargs="+",
                        help="""The min and max core-tail assignment constraints:
                        -1    : Train w/o assignment loss,
                        0     : Train with automatic assignment constraints
                        (c_min, c_max, t_min, t_max) : Train with user defined assignment constraints""")

    args = parser.parse_args()

    # Load data
    data_path = 'data/input_data/'
    agg_df = get_agg_data(args.target, args.data_name, data_path)

    train(
        agg_df,
        args.target_type,
        args.n_clusters,
        args.size_tuple,
        args.assign_tuple,
        args.batch_size,
        args.target,
        args.use_pretrain,
        args.lr,
        args.epochs,
        args.update_interval
    )
